@@ -1,21 +1,10 @@
 # region Backwards Compatibility
-
 from __future__ import (
     nested_scopes, generators, division, absolute_import, with_statement,
-   print_function, unicode_literals
+    print_function, unicode_literals
 )
 
 from sob.utilities.compatibility import backport
-backport()  # noqa
-
-import importlib
-
-try:
-    urlopen = importlib.import_module('urllib.request').urlopen
-except ImportError:
-    urlopen = importlib.import_module('urllib').urlopen
-
-# endregion
 
 import os
 import re
@@ -32,43 +21,41 @@ from sob.utilities import (
     get_source, property_name, properties_values, class_name,
     calling_function_qualified_name
 )
+from sob.utilities.compatibility import (
+    typing, backport, urlopen, urljoin, urlparse
+)
 from sob.model import (
     Model as ModelBase, Object, Dictionary, Array,
     from_meta as model_from_meta
 )
 from sob.utilities import qualified_name, url_relative_to
 from sob.properties import Property, Null
-
 from .oas.references import Resolver
 from .oas.model import OpenAPI, Schema, Reference, Parameter, Operation
 
-try:
-    _urllib_parse = importlib.import_module('urllib.parse')
-    urljoin = _urllib_parse.urljoin
-    urlparse = _urllib_parse.urlparse
-except ImportError:
-    _urllib = importlib.import_module('urllib')
-    urljoin = _urllib.urljoin
-    urlparse = _urllib.urlparse
-
-
-try:
-    from typing import Iterable, Sequence, Optional, Union
-except ImportError:
-    Iterable = Sequence = Optional = Union = None
-    _SchemaOrReference = None
+backport()
 
+Iterable = typing.Iterable
+Sequence = typing.Sequence
+Optional = typing.Optional
+Union = typing.Union
 
 # Constants
 
-
 _META_MODULE_QUALIFIED_NAME = qualified_name(meta)
 _META_PROPERTIES_QAULIFIED_NAME = qualified_name(meta.Properties)
 _META_PROPERTIES_QAULIFIED_NAME_LENGTH = len(_META_PROPERTIES_QAULIFIED_NAME)
-_DOC_POINTER_RE = re.compile(r'\s*([^\s](?:.|\r?\n)*[^\s])\s*\r?\n\s*\r?\n')
+_DOC_POINTER_RE = re.compile(
+    (
+        # Pointer
+        r'^(.*?)'
+        # Pointer stops at a double-return or end-of-string
+        r'(?:\r?\n\s*\r?\n|$)'
+    ),
+    re.DOTALL
+)
 _SPACES_RE = re.compile(r'[\s\n]')
 
-
 # region Functions
 
 
@@ -134,6 +121,19 @@ def operation_defines_model(operation):
 # region Private Classes
 
 
+class _RecursiveReferencePlaceholder(object):
+
+    def __init__(self, relative_url_pointer):
+        # type: (str) -> None
+        self.relative_url_pointer = relative_url_pointer
+
+    def __eq__(self, other):
+        # type: (_RecursiveReferencePlaceholder) -> bool
+        return (
+            self.relative_url_pointer == other.relative_url_pointer
+        )
+
+
 class _Modeler(object):
     """
     This class parses an OpenAPI schema and produces a data model based on the
@@ -168,7 +168,7 @@ class _Modeler(object):
         self.pointers_models = OrderedDict()
 
     def get_relative_url_pointer_model(self, relative_url_pointer):
-        # type: (str) -> Model
+        # type: (str) -> Module
         return self._relative_urls_pointers_models[relative_url_pointer]
 
     def set_relative_url_pointer_class_name(
@@ -190,11 +190,15 @@ class _Modeler(object):
         ] = relative_url_pointer
         return class_name_
 
-    def relative_url_pointer_exists(self, relative_url_pointer):
+    def relative_url_pointer_class_name_exists(self, relative_url_pointer):
         # type: (str) -> bool
         return relative_url_pointer in self._relative_urls_pointers_class_names
 
-    def class_name_exists(self, class_name_):
+    def relative_url_pointer_model_exists(self, relative_url_pointer):
+        # type: (str) -> bool
+        return relative_url_pointer in self._relative_urls_pointers_models
+
+    def class_name_relative_urls_pointer_exists(self, class_name_):
         # type: (str) -> bool
         return class_name_ in self._class_names_relative_urls_pointers
 
@@ -213,7 +217,8 @@ class _Modeler(object):
                 property_.types = []
             for schema in schemas:
                 child_property_ = self.get_property(schema)
-                property_.types.append(child_property_)
+                if child_property_ is not None:
+                    property_.types.append(child_property_)
 
     def property_all_of(self, property_, schemas):
         # type: (Iterable[Schema]) -> None
@@ -225,7 +230,8 @@ class _Modeler(object):
                 property_.types = []
             for schema in schemas:
                 child_property_ = self.get_property(schema)
-                property_.types.append(child_property_)
+                if child_property_ is not None:
+                    property_.types.append(child_property_)
 
     def get_property(
         self,
@@ -247,9 +253,12 @@ class _Modeler(object):
             self.property_all_of(property_, schema.all_of)
         elif schema_defines_model(schema):
             property_ = properties.Property()
-            property_.types = (
-                self.get_model(schema, is_property=(not is_referenced)),
+            property_types = self.get_model(
+                schema,
+                is_property=(not is_referenced)
             )
+            if property_types is not None:
+                property_.types = (property_types,)
         elif schema.type_ == 'number':
             property_ = properties.Number()
         elif schema.type_ == 'integer':
@@ -268,7 +277,9 @@ class _Modeler(object):
         elif schema.type_ == 'file':
             property_ = properties.Bytes()
         else:
-            raise ValueError(schema.type_)
+            raise ValueError(
+                'No schema "type" found:\n' + repr(schema)
+            )
         if schema.enum:
             property_ = properties.Enumerated(
                 values=tuple(schema.enum),
@@ -333,18 +344,30 @@ class _Modeler(object):
         relative_url, pointer = self.get_definition_relative_url_and_pointer(
             definition
         )
-
         return relative_url + pointer
 
+    def set_relative_url_pointer_model(self, relative_url_pointer, model):
+        # type: (str, Optional[sob.model.Model]) -> None
+        self._relative_urls_pointers_models[relative_url_pointer] = model
+
+    def set_model_class_name(self, model, class_name_=None):
+        # type: (Optional[sob.model.Model], Optional[str]) -> None
+        if class_name_ is None:
+            class_name_ = model.__name__
+        self._class_names_models[class_name_] = model
+
     def get_model(self, definition, is_property=False):
         # type: (Union[Schema, Operation, Parameter], bool) -> ModelBase
         relative_url_pointer = self.get_definition_relative_url_pointer(
             definition
         )
         # If this model has already been generated--use the cached model
-        if relative_url_pointer not in self._relative_urls_pointers_models:
-            # Setting this to `None` prevents recursion errors
-            self._relative_urls_pointers_models[relative_url_pointer] = None
+        if not self.relative_url_pointer_model_exists(relative_url_pointer):
+            # Setting this to prevents recursion errors
+            self.set_relative_url_pointer_model(
+                relative_url_pointer,
+                None
+            )
             if isinstance(definition, Schema):
                 model = self.get_schema_model(
                     definition,
@@ -374,9 +397,12 @@ class _Modeler(object):
                         repr(definition)
                     )
                 )
-            self._relative_urls_pointers_models[relative_url_pointer] = model
-            self._class_names_models[model.__name__] = model
-        return self._relative_urls_pointers_models[relative_url_pointer]
+            self.set_relative_url_pointer_model(
+                relative_url_pointer,
+                model
+            )
+            self.set_model_class_name(model)
+        return self.get_relative_url_pointer_model(relative_url_pointer)
 
     @property
     def models(self):
@@ -385,9 +411,21 @@ class _Modeler(object):
         for definition in self.model_definitions:
             model = self.get_model(definition)
             if model.__name__ in models_names:
+                existing_model_meta = meta.read(models_names[model.__name__])
+                new_model_meta = meta.read(model)
                 # Ensure this is just a repeat use of the same model, and not
                 # a different model of the same name
-                assert models_names[model.__name__] == model
+                if existing_model_meta != new_model_meta:
+                    raise RuntimeError(
+                        'An attempt was made to define a model using an '
+                        'existing model name (`%s`)"\n\n'
+                        'Existing Module Metadata:\n\n%s\n\n'
+                        'New Module Metadata:\n\n%s' % (
+                            model.__name__,
+                            repr(existing_model_meta),
+                            repr(new_model_meta)
+                        )
+                    )
             elif model not in (Array, Dictionary):
                 models_names[model.__name__] = model
                 yield model
@@ -448,15 +486,22 @@ class _Modeler(object):
             )
         elif items_schemas is not None:
             items_schemas = (items_schemas,)
+        item_types = None
         # If item types are defined--create a class, otherwise--use the base
         # class
         if items_schemas:
             item_types = []
             for items_schema in items_schemas:
                 if schema_defines_model(items_schema):
-                    item_types.append(self.get_model(items_schema))
+                    items_model = self.get_model(items_schema)
                 else:
-                    item_types.append(self.get_property(items_schema))
+                    items_model = self.get_property(items_schema)
+                # Checking to make sure the model is not `None` is
+                # necessary because is some fringe cases (involving cyclic
+                # references) a model cannot be retrieved at this point
+                if items_model is not None:
+                    item_types.append(items_model)
+        if item_types:
             array_meta = meta.Array(item_types=item_types)
             # If the array allows a single type--base the array name on that
             # type in order to allow re-use
@@ -560,9 +605,13 @@ class _Modeler(object):
             object_meta.properties[property_name_] = self.get_property(
                 property_,
                 name=None if property_name_ == name_ else name_,
-                required=True if (
-                    schema.required and (name_ in schema.required)
-                ) else False
+                required=(
+                    True if (
+                        schema.required and (
+                            name_ in schema.required
+                        )
+                    ) else False
+                )
             )
         name = self.set_relative_url_pointer_class_name(
             relative_url_pointer, name
@@ -618,7 +667,7 @@ class _Modeler(object):
     ):
         # type: (str, str, Optional[bool]) -> str
         relative_url_pointer = relative_url + pointer
-        if not self.relative_url_pointer_exists(relative_url_pointer):
+        if not self.relative_url_pointer_class_name_exists(relative_url_pointer):
             parts = [
                 part.replace('~1', '/') for part in
                 pointer.lstrip('#/').split('/')
@@ -647,8 +696,8 @@ class _Modeler(object):
 
             relative_url_pointer_class_name = get_class_name(start_index)
             while start_index and (
-                self.class_name_exists(relative_url_pointer_class_name) and
-                self.get_class_name_relative_url_pointer(
+                    self.class_name_relative_urls_pointer_exists(relative_url_pointer_class_name) and
+                    self.get_class_name_relative_url_pointer(
                     relative_url_pointer_class_name
                 ) != relative_url_pointer
             ):
@@ -668,21 +717,12 @@ class _Modeler(object):
         `Schema`, `Parameter`, and `Operation`
         """
         self._traversed_relative_urls_pointers = set()
-        for definition in self._traverse_model_definitions(
-            self.root, (OpenAPI,)
+        for definition in tuple(
+            self._traverse_model_definitions(
+                self.root, (OpenAPI,)
+            )
         ):
             yield self._resolve(definition)
-            # definition = self._resolve(definition)
-            # relative_url_pointer = self.get_definition_relative_url_pointer(
-            #     definition
-            # )
-            # if relative_url_pointer not in (
-            #     self._traversed_relative_urls_pointers
-            # ):
-            #     self._traversed_relative_urls_pointers.add(
-            #         relative_url_pointer
-            #     )
-            #     yield definition
 
     def _resolve(self, model_instance, types=None):
         # type: (ModelBase, Optional[Sequence[type, Property]]) -> ModelBase
@@ -693,10 +733,26 @@ class _Modeler(object):
             if types is None:
                 types = (Schema,)
             url = meta.url(model_instance)
-            model_instance = self.resolver.get_document(url).resolve(
-                urljoin(meta.pointer(model_instance), model_instance.ref),
-                types
+            pointer = urljoin(
+                meta.pointer(model_instance),
+                model_instance.ref
             )
+            while isinstance(model_instance, Reference):
+                resolved_model_instance = self.resolver.get_document(
+                    url
+                ).resolve(
+                    pointer,
+                    types
+                )
+                if resolved_model_instance is model_instance:
+                    raise RuntimeError(
+                        '`Reference` instance is self-referential: ' +
+                        pointer
+                    )
+                else:
+                    model_instance = resolved_model_instance
+            if isinstance(model_instance, Reference):
+                raise RuntimeError(pointer)
         return model_instance
 
     def _add_traversed_definition(
@@ -870,7 +926,7 @@ class _ModuleParser(object):
     @staticmethod
     def _get_model_relative_url_pointer_and_class_name(model):
         # type: (type) -> str
-        match = _DOC_POINTER_RE.match(
+        match = _DOC_POINTER_RE.search(
             model.__doc__
         )
         relative_url_pointer = None
@@ -899,11 +955,11 @@ class _ModuleParser(object):
 # region Public Classes
 
 
-class Model(object):
+class Module(object):
     """
     This class parses an Open API document and outputs a module defining
     classes to represent each schema defined in the Open API document as a
-    subclass of `sob.model.Model`.
+    subclass of `sob.model.Module`.
 
     Initialization Parameters:
 
@@ -986,12 +1042,14 @@ class Model(object):
         return relative_url, pointer
 
     def model_from_pointer(self, url_pointer):
-        # type: (str) -> Model
+        # type: (str) -> Module
         """
         Return a model from a pointer
         """
         relative_url, pointer = self._get_relative_url_and_pointer(url_pointer)
-        return self._modeler.get_relative_url_pointer_model(relative_url + pointer)
+        return self._modeler.get_relative_url_pointer_model(
+            relative_url + pointer
+        )
 
     def save(self, path):
         # type: (str) -> None
@@ -1003,16 +1061,18 @@ class Model(object):
         defined by elements residing at the same location as the documented
         JSON pointer in the pre-existing classes' docstrings.
         """
-
         # Make sure that any matching, existing classes use the same names
         self._parse_existing_module(path)
-
         # Save the module
         with open(path, 'w') as model_io:
             model_io.write(str(self))
 
     def __eq__(self, other):
-        # type: (Model) -> bool
+        # type: (Module) -> bool
         return isinstance(other, self.__class__) and str(self) == str(other)
 
+
+# For backwards compatibility...
+Model = Module
+
 # endregion
\ No newline at end of file
