@@ -17,7 +17,7 @@ resolver = Resolver(open_api_document)
 resolver.dereference()
 ```
 """
-
+from sob.utilities.compatibility import backport, typing
 from collections import OrderedDict
 
 import collections
@@ -25,19 +25,13 @@ from itertools import chain
 from urllib import request
 from urllib.error import HTTPError
 from urllib.parse import urlparse, urljoin
+from warnings import warn
 
 from sob import meta
 
-
-from jsonpointer import resolve_pointer
+from jsonpointer import resolve_pointer, JsonPointerException
 from sob.utilities import qualified_name
 
-try:
-    import typing
-    from typing import Union, Any, Sequence, Dict
-except ImportError:
-    typing = Union = Any = Sequence = Dict = None
-
 from sob.model import Model, Array, Dictionary, Object, unmarshal, detect_format
 from sob.properties import Property
 from sob.properties.types import NoneType
@@ -46,6 +40,13 @@ from sob.errors import get_exception_text
 from .model import Reference, OpenAPI
 from ..errors import ReferenceLoopError
 
+backport()
+
+Optional = typing.Optional
+Union = typing.Union
+Sequence = typing.Sequence
+Dict = typing.Dict
+
 
 class _Document(object):
 
@@ -106,10 +107,11 @@ class _Document(object):
         """
         Recursively dereference this objects and all items/properties
         """
-
         try:
             if isinstance(model_instance, Object):
-                self.dereference_object_properties(model_instance, recursive=recursive)
+                self.dereference_object_properties(
+                    model_instance, recursive=recursive
+                )
             elif isinstance(
                 model_instance,
                 (Array, collections.Sequence, collections.Set)
@@ -117,9 +119,15 @@ class _Document(object):
                 model_instance,
                 (str, bytes)
             ):
-                self.dereference_array_items(model_instance, recursive=recursive)
-            elif isinstance(model_instance, (Dictionary, dict, OrderedDict)):
-                self.dereference_dictionary_values(model_instance, recursive=recursive)
+                self.dereference_array_items(
+                    model_instance, recursive=recursive
+                )
+            elif isinstance(
+                model_instance, (Dictionary, dict, OrderedDict)
+            ):
+                self.dereference_dictionary_values(
+                    model_instance, recursive=recursive
+                )
             else:
                 raise TypeError(
                     'The argument must be an instance of `%s`, not %s' % (
@@ -128,7 +136,9 @@ class _Document(object):
                     )
                 )
         except (ReferenceLoopError, RecursionError):
-            if not recursive:
+            if recursive:
+                warn(get_exception_text())
+            else:
                 raise
 
     def prevent_infinite_recursion(self, model_instance):
@@ -174,7 +184,11 @@ class _Document(object):
                 setattr(
                     object_,
                     property_name,
-                    self.resolve(value.ref, types=(property_,), dereference=recursive)
+                    self.resolve(
+                        value.ref,
+                        types=(property_,),
+                        dereference=recursive
+                    )
                 )
             elif recursive and isinstance(value, Model):
                 self.dereference(value, recursive=recursive)
@@ -242,7 +256,7 @@ class _Document(object):
         Return the object referenced by a pointer
         """
         if pointer in self.pointers:
-            # This catches
+            # This catches recursion errors
             if self.pointers[pointer] is None:
                 raise ReferenceLoopError(pointer)
         else:
@@ -252,8 +266,13 @@ class _Document(object):
                 resolved = resolve_pointer(self.root, pointer[1:])
                 # Cast the resolved reference as one of the given types
                 resolved = self.unmarshal_resolved_reference(
-                    resolved, self.url, pointer, types=types
+                    resolved,
+                    self.url,
+                    pointer,
+                    types=types
                 )
+                if resolved is None:
+                    raise RuntimeError()
             else:
                 # Resolve a reference from another Open API document
                 url, document_pointer = self.get_url_pointer(pointer)
@@ -331,7 +350,7 @@ class Resolver(object):
             self.documents['#'] = document
 
     def get_document(self, url):
-        # type: (str) -> Model
+        # type: (str) -> _Document
         """
         Retrieve a document by URL, or use the cached document if previously
         retrieved
